/*
 * Copyright (c) 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Jul 20, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.uvos.webauth;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.zip.Inflater;
import java.util.zip.InflaterOutputStream;

import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.xmlbeans.XmlException;
import org.codehaus.xfire.util.Base64;

import eu.unicore.samly2.SAMLConstants;

import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.exceptions.AuthenticationException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.engine.InternalAuthorization;
import pl.edu.icm.unicore.uvos.engine.SecurityContext;
import pl.edu.icm.unicore.uvos.idhelpers.X509IdentityHelper;
import pl.edu.icm.unicore.uvos.saml.SAMLAuthnUtil;
import pl.edu.icm.unicore.uvos.saml.SAMLErrorException;
import pl.edu.icm.unicore.uvos.server.idresolver.Authenticator;
import pl.edu.icm.unicore.uvos.server.saml.SAMLAuthImpl;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.UVOSFault;
import xmlbeans.org.oasis.saml2.protocol.AuthnRequestDocument;
import xmlbeans.org.oasis.saml2.protocol.AuthnRequestType;
import xmlbeans.org.oasis.saml2.protocol.ResponseDocument;
import xmlbeans.org.oasis.saml2.protocol.StatusType;


public class AuthnServlet extends HttpServlet
{
	private static final long serialVersionUID = 1L;
	
	//set by login.ftl to mark that user submitted login information. 
	//Values: TLS -> use data from authed TLS, email -> use submitted login&passwd
	public static final String REQ_LOGIN_MODE = "loginMode";
	public static final String REQ_EMAIL = "email";
	public static final String REQ_PASSWORD = "password";
	
	//set by servlet to mark that user has an authed TLS session. Value is client's DN.
	public static final String REQ_TLS_AUTHED = "tlsAuthedDN";
	//set by servlet to mark the mode: either "test" (no SAML request) or "saml"
	public static final String REQ_MODE = "operationMode";
	//set by servlet to mark the that last login attempt was unsuccessful.
	public static final String REQ_LOGIN_ERROR = "loginError";
	//set by servlet to count unsuccessful login attempts.
	public static final String SES_LOGIN_ATTEMPT = "loginAttempt";
	//set by servlet to keep SAML request during login.
	public static final String SES_SAML_REQ = "sessionSAMLReq";
	//set by servlet with logged identity
	public static final String REQ_LOGGED = "loggedAs";
	//standard SAML POST binding name - SAML request is encoded in this field
	public static final String REQ_SAML_REQUEST = "SAMLRequest";
	//standard SAML POST binding name - SAML response is encoded in this field
	public static final String REQ_SAML_RESPONSE = "SAMLResponse";
	//set by servlet with SAML requester URL
	public static final String SES_SAML_REQUESTER = "samlService";
	//set by servlet with requested identity format
	public static final String REQ_SAML_ID_FORMAT = "requestedFormat";
	//set by servlet to mark that login was OK but SAML response is with error.
	public static final String REQ_SAML_ERROR = "samlError";
	//relay state
	public static final String RELAY_STATE = "RelayState";
	
	private UVOSConnector uvosCon;
	
	public AuthnServlet()
	{
		uvosCon = new UVOSConnector();
	}
	
	@Override
	protected void doGet(HttpServletRequest req, HttpServletResponse resp)
		throws ServletException, IOException
	{
		String samlRequest = req.getParameter(REQ_SAML_REQUEST);
		if (samlRequest != null)
			serveSAMLRequest(samlRequest, req, resp);
		else
			showLoginPage("test", req, resp);
	}

	@Override
	protected void doPost(HttpServletRequest req, HttpServletResponse resp)
		throws ServletException, IOException
	{
		String loginMode = req.getParameter(REQ_LOGIN_MODE);
		String samlRequest = req.getParameter(REQ_SAML_REQUEST);
		
		//initial call -> SAML was posted
		if (samlRequest != null)
		{
			serveSAMLRequest(samlRequest, req, resp);
		} else if (loginMode != null)
		{ 
			//login data was submitted
			if (loginMode.equals("TLS"))
				servePostLoginTLS(req, resp);
			else
				servePostLoginEmail(req, resp);
		}
	}

	
	private void showLoginPage(String mode, 
			HttpServletRequest req, HttpServletResponse resp) 
		throws ServletException, IOException
	{
		HttpSession ses = req.getSession(true);
		if (mode != null)
			ses.setAttribute(REQ_MODE, mode);
		X509Certificate certFromTLS = checkTLS(req);
		if (certFromTLS != null)
			req.setAttribute(REQ_TLS_AUTHED, 
				certFromTLS.getSubjectX500Principal().getName());

		dispatchTo("login.ftl", req, resp);
	}

	private String inflateSAMLRequest(String samlRequest) throws IOException
	{
		//yes - this is buggy (platform encoding used). However no way to make it ok
		//as opensaml2 compresses using the platform encoding too.

		byte[] third = Base64.decode(samlRequest);
		Inflater decompressor = new Inflater(true);
		decompressor.setInput(third, 0, third.length);
		ByteArrayOutputStream baos = new ByteArrayOutputStream(1024);
		InflaterOutputStream os = new InflaterOutputStream(baos, decompressor);
		os.write(third);
		os.finish();
		os.close();
		return new String(baos.toByteArray());
	}

	
	private void serveSAMLRequest(String samlRequest, HttpServletRequest req, 
			HttpServletResponse resp) throws ServletException, IOException
	{
		String decodedReq = "";

		try
		{
			if (req.getMethod().equals("POST"))
				decodedReq = new String(Base64.decode(samlRequest));
			else if (req.getMethod().equals("GET"))
				decodedReq = inflateSAMLRequest(samlRequest);
		} catch (IOException e)
		{
			doError(req, resp, "Received a request which can't be translated into XML form: " + e);
			return;
		}
		AuthnRequestDocument reqDoc;
		try
		{
			reqDoc = AuthnRequestDocument.Factory.parse(decodedReq);
		} catch (XmlException e)
		{
			doError(req, resp, 
				"Received a nonparseable SAML request: " + e);
			return;
		}
		SAMLAuthImpl impl = uvosCon.getSamlAuthn();
		SAMLAuthnUtil samlUtil;
		try
		{
			samlUtil = impl.preprocessAuthnRequest(reqDoc);
		} catch (SAMLErrorException e)
		{
			doError(req, resp, "Received a wrong SAML request: " + e);
			return;
		}
		
		AuthnRequestType aReq = reqDoc.getAuthnRequest();
		if (aReq.getProtocolBinding() != null && 
				!aReq.getProtocolBinding().equals(
					"urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"))
		{
			doError(req, resp, "Received a wrong SAML request: " +
					"SAML binding required " +
					"by this request is not a HTTP-POST");
			return;
		}
		if (aReq.isSetIsPassive() && aReq.getIsPassive())
		{
			doError(req, resp, "Received a wrong SAML request: " +
					"passive authentication is required " +
					"by the request, while it is unsupported.");
			return;			
		}
		
		HttpSession ses = req.getSession();
		ses.setAttribute(SES_SAML_REQ, samlUtil);
		ses.setAttribute(SES_SAML_REQUESTER, samlUtil.getConsumerURL());
		ses.setAttribute(REQ_SAML_ID_FORMAT, samlUtil.getRequestedFormat());
		ses.removeAttribute(SES_LOGIN_ATTEMPT);
		String rs = req.getParameter(RELAY_STATE);
		if (rs != null)
			ses.setAttribute(RELAY_STATE, rs);
		showLoginPage("SAML", req, resp);
	}
	
	private void servePostLoginTLS(HttpServletRequest req, 
			HttpServletResponse resp) throws ServletException, IOException
	{
		X509Certificate cert = checkTLS(req);
		if (cert == null)
		{
			doError(req, resp, "TLS authentication was required but " +
					"there is no certificate available.");
			return;
		}
		String dn = cert.getSubjectX500Principal().getName();
		try
		{
			authAsDN(dn);
		} catch (Exception e)
		{
			try
			{
				authAsCert(cert);
			} catch (AuthenticationException e2)
			{
				authFail(req, resp);
				return;
			} catch (Exception e1)
			{
				doError(req, resp, e.toString());
				return;
			}
		}
		try
		{
			if (checkSAMLish(req))
				processSAML(req);
		} catch (UVOSFault e)
		{
			doError(req, resp, "Couldn't process SAML " +
					"authentication request: " + e);
			return;
		}
		authSuccess(req, resp);
	}

	private void servePostLoginEmail(HttpServletRequest req, 
			HttpServletResponse resp) throws ServletException, IOException
	{
		String passwd = req.getParameter(REQ_PASSWORD);
		String email = req.getParameter(REQ_EMAIL);
		try
		{
			authAsEmail(email, passwd);
		} catch (AuthenticationException e)
		{
			authFail(req, resp);
			return;
		} catch (Exception e)
		{
			doError(req, resp, e.toString());
			return;
		}
		
		try
		{
			if (checkSAMLish(req))
				processSAML(req);
		} catch (UVOSFault e)
		{
			doError(req, resp, "Couldn't process SAML " +
					"authentication request: " + e);
			return;
		}

		authSuccess(req, resp);
	}

	private void processSAML(HttpServletRequest req) 
		throws ServletException, IOException, UVOSFault
	{
		HttpSession ses = req.getSession();
		SAMLAuthnUtil samlUtil = (SAMLAuthnUtil) ses.getAttribute(SES_SAML_REQ);
		
		SAMLAuthImpl impl = uvosCon.getSamlAuthn();
		ResponseDocument respDoc; 
		respDoc = impl.processAuthnRequest(
				samlUtil, impl.setupAuthnContext(),
				Calendar.getInstance());
		String assertion = respDoc.xmlText();
		String encodedAssertion = Base64.encode(assertion.getBytes());
		req.setAttribute(REQ_SAML_RESPONSE, encodedAssertion);	
		
		StatusType status = respDoc.getResponse().getStatus();
		if (status == null || status.getStatusCode() == null)
			req.setAttribute(REQ_SAML_ERROR, "Severe SAML processing error.");
		if (!status.getStatusCode().getValue().equals(SAMLConstants.STATUS_OK))
			req.setAttribute(REQ_SAML_ERROR, status.getStatusMessage());
	}
	
	private void authAsEmail(String email, String passwd) 
		throws AuthenticationException, InternalException, InvalidValueException
	{
		Identity id = new Identity(IdentityType.EMAIL, 
				email, null);
		id.setAuthnData(passwd.equals("") ? null : passwd);
		authenticate(id);
	}

	private void authAsDN(String dn) 
		throws AuthenticationException, InternalException, InvalidValueException
	{
		Identity id = new Identity(IdentityType.DN, 
				dn, null);
		authenticate(id);
	}

	private void authAsCert(X509Certificate cert) 
		throws AuthenticationException, InternalException,
		InvalidValueException, CertificateException
	{
		Identity id = X509IdentityHelper.convertCert(cert);
		authenticate(id);
	}

	private void authenticate(Identity id) 
		throws AuthenticationException, InternalException
	{
		InternalAuthorization voAuthz = uvosCon.getAuthzModule();
		try
		{
			SecurityContext ret = voAuthz.establishSecurityContext(id);
			Authenticator.injectFakeContext(ret);
		} catch (IdentityNotKnownException e)
		{
			throw new AuthenticationException("Unknown principial");
		} catch (InvalidValueException e)
		{
			throw new AuthenticationException("Unknown principial " +
					"(wrong format: " + e.getMessage() + ")");
		} 
	}

	private X509Certificate checkTLS(HttpServletRequest req)
	{
		X509Certificate[] certs = (X509Certificate[]) 
			req.getAttribute("javax.servlet.request.X509Certificate");
		if (certs != null && certs.length > 0)
			return certs[0];
		return null;
	}
	
	private boolean checkSAMLish(HttpServletRequest req)
	{
		HttpSession ses = req.getSession();
		if (ses != null)
		{
			String mode = (String)ses.getAttribute(REQ_MODE);
			return mode != null && mode.equals("SAML");
		}
		return false;
	}
	
	private void authFail(HttpServletRequest req, 
			HttpServletResponse resp) throws ServletException, IOException
	{
		HttpSession session = req.getSession(true);
		Integer attempt = (Integer) session.getAttribute(SES_LOGIN_ATTEMPT);
		if (attempt == null)
			attempt = 1;
		else
			attempt++;
		if (attempt > 2)
		{
			doError(req, resp, "Too many unsuccessful attempts were performed.");
			return;
		}
		req.setAttribute(REQ_LOGIN_ERROR, "Authentication failed");
		session.setAttribute(SES_LOGIN_ATTEMPT, attempt);
		showLoginPage(null, req, resp);
	}

	private void authSuccess(HttpServletRequest req, HttpServletResponse resp) 
		throws ServletException, IOException
	{
		String id;
		try
		{
			id = Authenticator.getCurrentContext().getCaller().prettyString();
		} catch (InternalException e)
		{
			throw new ServletException("Can't retrieve authentication " +
					"context, internal implementation problem", e);
		}
		req.setAttribute(REQ_LOGGED, id);
		dispatchTo("success.ftl", req, resp);
	}
	
	private void doError(HttpServletRequest req, 
			HttpServletResponse resp, String msg) 
		throws ServletException, IOException
	{
		req.setAttribute(REQ_LOGIN_ERROR, "Server error: " + msg);
		dispatchTo("failure.ftl", req, resp);
	}
	
	
	private void dispatchTo(String url, HttpServletRequest req, 
			HttpServletResponse resp) 
		throws ServletException, IOException
	{
		RequestDispatcher dispatcher = req.getRequestDispatcher(url);
		if (dispatcher != null)
			dispatcher.forward(req, resp);		
	}
}
